"An opbject for a Vertex"
Object subclass: #Vertex
	   instanceVariableNames: 'x y z w'
	   classVariableNames: ''
	   poolDictionaries: ''
	   category: 'OpenGL' !

Vertex comment: 
'My instances are vertexes in either a 2D or 3D space. I also have
another axe for matrix works' !

!Vertex class methodsFor: 'creation'!
x: anX y: aY
	^ self new x: anX y: aY!
x: anX y: aY z: aZ
	^ self new x: anX y: aY z: aZ!
x: anX y: aY z: aZ w: aW
	^ self new x: anX y: aY z: aZ w: aW!
!

!Vertex methodsFor: 'accessing'!
x
	^x!
y
	^y!
z
	^z!
w
	^w!

x: anX
	x := anX!

y: aY
	y := aY!

z: aZ
	z := aZ!

w: aW
	w := aW!

x: anX y: aY
	x := anX.
	y := aY!

x: anX y: aY z: aZ
	x := anX.
	y := aY.
	z := aZ!

x: anX y: aY z: aZ w: aW
	x := anX.
	y := aY.
	z := aZ.
	w := aW.!
!


!Vertex methodsFor: 'Transforming'!
asArray
	| anArray |
	z isNil 
		ifTrue: [anArray := Array with: x with: y.]
		ifFalse: [
			w isNil 
				ifTrue: [anArray := Array with: x with: y with: z.]
				ifFalse: [anArray := Array with: x with: y with: z with: w.].
			].
	^anArray !
!

"An object for a point (exactly the same as vertex...)"
Vertex subclass: #Point
	   instanceVariableNames: ''
	   classVariableNames: ''
	   poolDictionaries: ''
	   category: 'OpenGL' !

Point comment: 
'My instances are point in either a 2D or 3D space. I also have
another axe for either transparency or matrix works' !

!Point class methodsFor: 'creation'!
x: anX y: aY
	^ self new x: anX y: aY!
x: anX y: aY z: aZ
	^ self new x: anX y: aY z: aZ!
x: anX y: aY z: aZ w: aW
	^ self new x: anX y: aY z: aZ w: aW!
!


" An object that represents colors "
Object subclass: #Color
	   instanceVariableNames: 'red green blue transparency'
	   classVariableNames: ''
	   poolDictionaries: ''
	   category: 'OpenGL' !

Color comment: 
'My instances are color represented as a combinasion of Red, Green, blue and Transparency'!


!Color methodsFor: 'Accessing'!
red: aRed
	red := aRed!
green: aGreen
	green := aGreen!
blue: aBlue
	blue := aBlue!
transparency: aTransparency
	transparency := aTransparency!
red
	^red!
green
	^green!
blue
	^blue!
transparency
	^transparency! !
"asMatrix
	^#(red green blue)! !"

!Color class methodsFor: 'main colors'!
black
	^ self new red: 0.0 ; green: 0.0 ; blue: 0.0 ; transparency: 0.0 !
white
	^ self new red: 1.0 ; green: 1.0 ; blue: 1.0 ; transparency: 0.0 !
red
	^ self new red: 1.0 ; green: 0.0 ; blue: 0.0 ; transparency: 0.0 !
green
	^ self new red: 0.0 ; green: 1.0 ; blue: 0.0 ; transparency: 0.0 !
blue
	^ self new red: 0.0 ; green: 0.0 ; blue: 1.0 ; transparency: 0.0 !
red: aRed green: aGreen blue: aBlue transparency: aTransparency
	^ self new red: aRed ; green: aGreen ; blue: aBlue ; transparency: aTransparency !
red: aRed green: aGreen blue: aBlue 
	^ self new red: aRed ; green: aGreen ; blue: aBlue ; transparency: 1.0 !
!

!Color methodsFor: 'Transforming'!
asArray
	| anArray |
	transparency isNil 
		ifTrue:  [anArray := Array with: red with: green with: blue.]
		ifFalse: [anArray := Array with: red with: green with: blue with: transparency.].
	^anArray !
!



" OpenGL class "
Object subclass: #OpenGL
	   instanceVariableNames: 'callbacks'
	   classVariableNames: 'init'
	   poolDictionaries: ''
	   category: 'OpenGL' !
OpenGL comment:
'My instances are interfaces to OpenGL library with access to GL, Glu and Glut.
See OpenGL programming guide for more informations.'!




" ******** "
" GL funcs "
" ******** "

!OpenGL methodsFor: 'GL'!

glClearColor: aColor
	self cacheGlClearColor: aColor red green: aColor green blue: aColor blue !

cacheGlClearColor: aRed green: aGreen blue: aBlue
	<cCall: 'glClearColor' returning: #void args: #( #float #float #float )> !

glShadeModel: aModel
	<cCall: 'glShadeModel' returning: #void args: #( #int )> !

glClear: aClear
	<cCall: 'glClear' returning: #void args: #( #int )> !

glColor3f: aColor
	self glColor3f: aColor red green: aColor green blue: aColor blue !

glColor3f: aRed green: aGreen blue: aBlue
	<cCall: 'glColor3f' returning: #void args: #( #float #float #float )> !

glLoadIdentity
	<cCall: 'glLoadIdentity' returning: #void args: #()> !

glScalef: aVertex
	self glScalef: aVertex x y: aVertex y z: aVertex z !

glScalef: anX y: anY z: anZ
	<cCall: 'glScalef' returning: #void args: #( #float #float #float )> !

glFlush
	<cCall: 'glFlush' returning: #void args: #()> !

glFrustum: left right: right bottom: bottom top: top near: near far: far
	<cCall: 'glFrustum' returning: #void args: #( #double #double #double #double #double #double )> !

glOrtho: left right: right bottom: bottom top: top near: near far: far
	<cCall: 'glOrtho' returning: #void args: #( #double #double #double #double #double #double )> !

glMatrixMode: mode
	<cCall: 'glMatrixMode' returning: #void args: #( #int )> !

glViewport: aPoint extend: extend
	self cachedGlViewport: aPoint x y: aPoint y width: extend x height: extend y!

cachedGlViewport: x y: y width: width height: height
	<cCall: 'glViewport' returning: #void args: #(#int #int #int #int)>!

glRotatef: angle direction: aVertex
	^ self glRotatef: angle x: aVertex x y: aVertex y z: aVertex z!

glRotatef: angle x: anX y: aY z: aZ
	<cCall: 'glRotatef' returning: #void args: #( #float #float #float #float)>!

glRotated: angle x: anX y: aY z: aZ
	<cCall: 'glRotated' returning: #void args: #( #double #double #double #double)>!

glTranslatef: aVertex
	self glTranslatef: aVertex x y: aVertex y z: aVertex z!

glTranslatef: x y: y z: z
	<cCall: 'glTranslatef' returning: #void args: #( #float #float #float)>!

glTranslated: aVertex
	self glTranslated: aVertex x y: aVertex y z: aVertex z!

glTranslated: aX y: aY z: aZ
	<cCall: 'glTranslated' returning: #void args: #( #int #int #int )>!

glLightPosition: number position: aPoint
	| anArray size |
	anArray := Array new: 4.
	aPoint asArray doWithIndex: [:value :index | 
		anArray at: index put: value].
	size := aPoint asArray size.
	size < 4 ifTrue: [
		anArray replaceFrom: size to: 4 withObject: 0.0.
		].
	self glLightfv: number 
		 property: OpenGL GlPosition 
		 value: anArray.!

glLightfv: lightNumber property: aProp value: anArray
	<cCall: 'glLightfv' returning: #void args: #( #int #int #smalltalk )>!

glEnable: aProperty
	<cCall: 'glEnable' returning: #void args: #( #int )> !

glDisable: aProperty
	<cCall: 'glDisable' returning: #void args: #( #int )> !

glColorMaterial: aFace mode: aMode
	<cCall: 'glColorMaterial' returning: #void args: #( #int #int )>!

glMaterialf: aFace mode: aMode value: aValue
	<cCall: 'glMaterialf' returning: #void args: #( #int #int #float )>!

glMaterialfv: aFace mode: aMode value: anArray	
	Transcript show: anArray printString , ' (' , anArray size printString, ')' ; cr.
	self glMaterialfv: aFace mode: aMode value: anArray!

glMaterialfv: aFace mode: aMode value: anArray
	<cCall: 'glMaterialfv' returning: #void args: #( #int #int #smalltalk )>!

glMap1f: target u1: u1 u2: u2 ustride: uStride uorder: uOrdrer points: arrayOfPoints
	<cCall: 'glMap2f' returning: #void args: #(#int #float #float #int #int #smalltalk)>!

glMap1d: target u1: u1 u2: u2 ustride: uStride uorder: uOrdrer points: arrayOfPoints
	<cCall: 'glMap2f' returning: #void args: #(#int #double #double #int #int #smalltalk)>!

glMap2f: target u1: u1 u2: u2 ustride: uStride uorder: uOrdrer v1: v1 v2: v2 vstride: vStride vorder: vOrder points: arrayOfPoints
	<cCall: 'glMap2f' returning: #void args: #(#int #float #float #int #int #float #float #int #int #smalltalk)>!

glMap2d: target u1: u1 u2: u2 ustride: uStride uorder: uOrdrer v1: v1 v2: v2 vstride: vStride vorder: vOrder points: arrayOfPoints
	<cCall: 'glMap2f' returning: #void args: #(#int #double #double #int #int #double #double #int #int #smalltalk)>!

glMapGrid2f: nu u1: u1 u2: u2 nv: nv v1: v1 v2: v2
	<cCall: 'glMapGrid2f' returning: #void args: #(#int #float #float #int #float #float )>!

glMapGrid2d: nu u1: u1 u2: u2 nv: nv v1: v1 v2: v2
	<cCall: 'glMapGrid2f' returning: #void args: #(#int #double #double #int #double #double )>!

glEvalCoord2f: aX y: aY
	<cCall: 'glEvalCoord2f' returning: #void args: #( #float #float )> !

glBegin: type
	<cCall: 'glBegin' returning: #void args: #(#int)>!

glEnd
	<cCall: 'glEnd' returning: #void args: #()> !

glPushMatrix
	<cCall: 'glPushMatrix' returning: #void args: #( )> !

glPopMatrix
	<cCall: 'glPopMatrix' returning: #void args: #( )> !

glPointSize: size
	<cCall: 'glPointSize' returning: #void args: #(#float)> !

glVertex3f: aPoint
	self glVertex3f: aPoint x y: aPoint y z: aPoint z !

glVertex3f: aX y: aY z: aZ
	<cCall: 'glVertex3f' returning: #void args: #(#float #float #float)>!


" * * * * * * "
" Display list"
" * * * * * * "

glIsList: listNum
	<cCall: 'glIsList' returning: #boolean args: #( #int ) >!

glDeleteLists: listNum range: range
	<cCall: 'glDeleteLists' returning: #void args: #( #int #int ) >!

glGenLists: range
	<cCall: 'glGenLists' returning: #int args: #( #int )>!

glNewList: list mode: mode
	<cCall: 'glNewList' returning: #void args: #( #int #int ) >!

glEndList
	<cCall: 'glEndList' returning: #void args: #( ) >!

glCallList: list
	<cCall: 'glCallList' returning: #void args: #( #int ) >!

glCallLists: n type: type lists: lists
	<cCall: 'glCallLists' returning: #void args: #( #int #int #smalltalk ) >!

glListBase: base
	<cCall: 'glListBase' returning: #void args: #( #int ) >!

!



" *********** "
" OpenGL Main "
" *********** "


!OpenGL class methodsFor: 'Main interactions'!
mainIteration
	'Forking  main interaction...' printNl.
	^([self realMainIteration] newProcess)
        priority: Processor userBackgroundPriority;
        name: 'OpenGLMainIteration';
        resume;
        yourself !


realMainIteration
    <cCall: 'glutMainLoop' returning: #void args: #( )>!
!

!OpenGL methodsFor: 'Main interactions'!
connect: event 
	<cCall: 'openglConnectSignal' 
		returning: #void
		args: #( #selfSmalltalk  #int )>!

!
!OpenGL methodsFor: 'signal'!
signal
	'I received a signal...' printNl!
!

