#include "config.h"
#include "gstpub.h"

#include <stdio.h>
#include <GL/glut.h>


static VMProxy* _opengl_vm_proxy = NULL ;
#define nil  _opengl_vm_proxy->nilOOP

#define DEBUG 1

#ifdef DEBUG
static FILE* fp = NULL ;
#endif


/* * * * * * */
/* the NURBS */
/* callbacks */
/* * * * * * */



/* Number of known callbacks */
#define NbGluNurbFunctions 13
/* Receivers for the callbacks; initialized in module init func */
static OOP gluNurbReceiver[NbGluNurbFunctions] ;
/* Selectors; initialized in module init func */
static OOP gluNurbSelector[NbGluNurbFunctions] ;
/* Callback constants definition */
static GLenum gluNurbFunctions[NbGluNurbFunctions] = {
  GLU_NURBS_ERROR,
  GLU_NURBS_BEGIN,
  GLU_NURBS_VERTEX,
  GLU_NURBS_NORMAL,
  GLU_NURBS_COLOR,
  GLU_NURBS_TEXTURE_COORD,
  GLU_NURBS_END,
  GLU_NURBS_BEGIN_DATA,
  GLU_NURBS_VERTEX_DATA,
  GLU_NURBS_NORMAL_DATA,
  GLU_NURBS_COLOR_DATA,
  GLU_NURBS_TEXTURE_COORD_DATA,
  GLU_NURBS_END_DATA
} ;


/* Retreive the callback index from the Callback contant definition */
int gst_glu_nurbCallback_getCallbackIndex(GLenum which)
{
  int i ;
  for (i = 0 ; i < NbGluNurbFunctions ; ++i) {
	/* If function is found, return index */
	if (gluNurbFunctions[i] == which) {
	  return i ;
	}
  }
  return -1 ;
}

/* Send the callback message to the receiver */
void gst_glu_nurbCallback_sendMessage(GLenum which, int nArgs, OOP* args)
{
  /* Get the callback index */
  int index = gst_glu_nurbCallback_getCallbackIndex(which) ;
  if(index == -1) return ;
  /* verify the receiver and the selector */
  if(gluNurbReceiver[index] != nil && 
	 gluNurbSelector[index] != nil) 
	{
	  /* Send the callback message */
	  if(nArgs > 0) 
		_opengl_vm_proxy->nvmsgSend(gluNurbReceiver[index], 
									gluNurbSelector[index],
									args, nArgs) ;
	  else
		_opengl_vm_proxy->perform(gluNurbReceiver[index], 
								  gluNurbSelector[index]) ;
	}
}
/* Define the C Callback functions and call the sendMessage function */
void gst_glu_nurbCallback_Nurbs_Error (GLenum errorCode)
{
  OOP* args = alloca ( sizeof(OOP) * 1) ;
  args[0] = _opengl_vm_proxy->intToOOP(errorCode) ;
  gst_glu_nurbCallback_sendMessage(GLU_NURBS_ERROR, 1, args) ;
}

void gst_glu_nurbCallback_Nurbs_Begin ( GLenum type )
{
  OOP* args = alloca ( sizeof(OOP) * 1) ;
  args[0] = _opengl_vm_proxy->intToOOP(type) ;
  gst_glu_nurbCallback_sendMessage(GLU_NURBS_BEGIN, 1, args) ;
}

void gst_glu_nurbCallback_Nurbs_Vertex ( GLfloat* vertex )
{
  OOP* args = alloca ( sizeof(OOP) * 3 ) ;
  args[0] = _opengl_vm_proxy->floatToOOP( vertex[0] ) ;
  args[1] = _opengl_vm_proxy->floatToOOP( vertex[1] ) ;
  args[2] = _opengl_vm_proxy->floatToOOP( vertex[2] ) ;
  gst_glu_nurbCallback_sendMessage(GLU_NURBS_VERTEX, 3, args) ;
}

void gst_glu_nurbCallback_Nurbs_Normal ( GLfloat* normal )
{
  OOP* args = alloca ( sizeof(OOP) * 3 ) ;
  args[0] = _opengl_vm_proxy->floatToOOP( normal[0] ) ;
  args[1] = _opengl_vm_proxy->floatToOOP( normal[1] ) ;
  args[2] = _opengl_vm_proxy->floatToOOP( normal[2] ) ;
  gst_glu_nurbCallback_sendMessage(GLU_NURBS_NORMAL, 3, args) ;
}

void gst_glu_nurbCallback_Nurbs_Color ( GLfloat* color)
{
  OOP* args = alloca ( sizeof(OOP) * 3 ) ;
  args[0] = _opengl_vm_proxy->floatToOOP( color[0] ) ;
  args[1] = _opengl_vm_proxy->floatToOOP( color[1] ) ;
  args[2] = _opengl_vm_proxy->floatToOOP( color[2] ) ;
  gst_glu_nurbCallback_sendMessage(GLU_NURBS_COLOR, 3, args) ;
}

void gst_glu_nurbCallback_Nurbs_Texture_Coord ( GLfloat* tCrd )
{
  OOP* args = alloca ( sizeof(OOP) * 3 ) ;
  args[0] = _opengl_vm_proxy->floatToOOP( tCrd[0] ) ;
  args[1] = _opengl_vm_proxy->floatToOOP( tCrd[1] ) ;
  args[2] = _opengl_vm_proxy->floatToOOP( tCrd[2] ) ;
  gst_glu_nurbCallback_sendMessage(GLU_NURBS_TEXTURE_COORD, 3, args) ;
}

void gst_glu_nurbCallback_Nurbs_End ( )
{
  gst_glu_nurbCallback_sendMessage(GLU_NURBS_END, 0, &nil) ;
}

void gst_glu_nurbCallback_Nurbs_Begin_Data ( GLenum type, void* data )
{
  OOP* args = alloca ( sizeof(OOP) * 2) ;
  args[0] = _opengl_vm_proxy->intToOOP(type) ;
  args[1] = (OOP) data ;
  gst_glu_nurbCallback_sendMessage(GLU_NURBS_BEGIN_DATA, 2, args) ;
}

void gst_glu_nurbCallback_Nurbs_Vertex_Data ( GLfloat* vertex, void* data )
{
  OOP* args = alloca ( sizeof(OOP) * 4 ) ;
  args[0] = _opengl_vm_proxy->floatToOOP( vertex[0] ) ;
  args[1] = _opengl_vm_proxy->floatToOOP( vertex[1] ) ;
  args[2] = _opengl_vm_proxy->floatToOOP( vertex[2] ) ;
  args[3] = (OOP) data ;
  gst_glu_nurbCallback_sendMessage(GLU_NURBS_VERTEX_DATA, 4, args) ;
}

void gst_glu_nurbCallback_Nurbs_Normal_Data ( GLfloat* normal, void* data )
{
  OOP* args = alloca ( sizeof(OOP) * 4 ) ;
  args[0] = _opengl_vm_proxy->floatToOOP( normal[0] ) ;
  args[1] = _opengl_vm_proxy->floatToOOP( normal[1] ) ;
  args[2] = _opengl_vm_proxy->floatToOOP( normal[2] ) ;
  args[3] = (OOP) data ;
  gst_glu_nurbCallback_sendMessage(GLU_NURBS_NORMAL_DATA, 4, args) ;
}

void gst_glu_nurbCallback_Nurbs_Color_Data ( GLfloat* color, void* data)
{
  OOP* args = alloca ( sizeof(OOP) * 4 ) ;
  args[0] = _opengl_vm_proxy->floatToOOP( color[0] ) ;
  args[1] = _opengl_vm_proxy->floatToOOP( color[1] ) ;
  args[2] = _opengl_vm_proxy->floatToOOP( color[2] ) ;
  args[3] = (OOP) data ;
  gst_glu_nurbCallback_sendMessage(GLU_NURBS_COLOR_DATA, 4, args) ;
}

void gst_glu_nurbCallback_Nurbs_Texture_Coord_Data ( GLfloat* tCrd, void* data )
{
  OOP* args = alloca ( sizeof(OOP) * 4 ) ;
  args[0] = _opengl_vm_proxy->floatToOOP( tCrd[0] ) ;
  args[1] = _opengl_vm_proxy->floatToOOP( tCrd[1] ) ;
  args[2] = _opengl_vm_proxy->floatToOOP( tCrd[2] ) ;
  args[3] = (OOP) data ;
  gst_glu_nurbCallback_sendMessage(GLU_NURBS_TEXTURE_COORD_DATA, 4, args) ;
}

void gst_glu_nurbCallback_Nurbs_End_Data (void* data )
{
  OOP* args = alloca ( sizeof(OOP) * 1 ) ;
  args[0] = (OOP) data ;
  gst_glu_nurbCallback_sendMessage(GLU_NURBS_END_DATA, 1, args) ;
}


/* Pointers to the Glu Nurbs callback functions */
static _GLUfuncptr gst_glu_nurbCallback[NbGluNurbFunctions] = {
  gst_glu_nurbCallback_Nurbs_Error,
  gst_glu_nurbCallback_Nurbs_Begin,
  gst_glu_nurbCallback_Nurbs_Vertex,
  gst_glu_nurbCallback_Nurbs_Normal,
  gst_glu_nurbCallback_Nurbs_Color,
  gst_glu_nurbCallback_Nurbs_Texture_Coord,
  gst_glu_nurbCallback_Nurbs_End,
  gst_glu_nurbCallback_Nurbs_Begin_Data,
  gst_glu_nurbCallback_Nurbs_Vertex_Data,
  gst_glu_nurbCallback_Nurbs_Normal_Data,
  gst_glu_nurbCallback_Nurbs_Color_Data,
  gst_glu_nurbCallback_Nurbs_Texture_Coord_Data,
  gst_glu_nurbCallback_Nurbs_End_Data
} ;



void gst_opengl_gluNurbsCallback (OOP nurb, GLenum which, OOP selector, OOP receiver)
{ 
  int index = gst_glu_nurbCallback_getCallbackIndex(which) ;
  if(index == -1) 
	return ;

  /* Unregister previous receiver and selector callback */
  if(gluNurbReceiver[index] != nil) 
	{
	  _opengl_vm_proxy->unregisterOOP(gluNurbReceiver[index]) ;
	  gluNurbReceiver[index] = nil ;
	}
  if(gluNurbSelector[index] != nil) 
	{
	  _opengl_vm_proxy->unregisterOOP(gluNurbSelector[index]) ;
	  gluNurbSelector[index] = nil ;
	}
  /* Verify if selector and receiver are not nil. Else clear the callback */
  if(receiver != nil && selector != nil ) 
	{
	  /* Assign new receiver and selector and register them*/
	  gluNurbReceiver[index] = receiver ;
	  gluNurbSelector[index] = selector ;
	  _opengl_vm_proxy->registerOOP(gluNurbReceiver[index]) ;
	  _opengl_vm_proxy->registerOOP(gluNurbSelector[index]) ;

	  /* Setting up the callback pointer */
	  gluNurbsCallback (_opengl_vm_proxy->OOPToCObject(nurb), which, gst_glu_nurbCallback[index]) ;
	}
  else 
	{
	  gluNurbReceiver[index] = nil ;
	  gluNurbSelector[index] = nil ;

	  /* Cleaning up the callback */
	  gluNurbsCallback (_opengl_vm_proxy->OOPToCObject(nurb), which, NULL) ;
	}
}

void gst_opengl_gluNurbsCallbackData (OOP nurb, OOP userData)
{
  gluNurbsCallbackData (_opengl_vm_proxy->OOPToCObject(nurb),
						(void*)userData) ;
}
void gst_opengl_gluNurbsCallbackDataEXT (OOP nurb, OOP userData)
{
  gluNurbsCallbackDataEXT (_opengl_vm_proxy->OOPToCObject(nurb),
						   (void*)userData) ;
}

void gst_opengl_gluDeleteNurbsRenderer (OOP nurb)
{
  gluDeleteNurbsRenderer (_opengl_vm_proxy->OOPToCObject(nurb)) ;
}
GLfloat gst_opengl_gluGetNurbsProperty (OOP nurb, GLenum property)
{
  GLfloat data ;
  gluGetNurbsProperty (_opengl_vm_proxy->OOPToCObject(nurb), property, &data) ;
  return data ;
}
void gst_opengl_gluNurbsCurve (OOP nurb, GLint knotCount, GLfloat *knots, GLint stride, GLfloat *control, GLint order, GLenum type)
{
  gluNurbsCurve (_opengl_vm_proxy->OOPToCObject(nurb),
				 knotCount, 
				 knots, 
				 stride, 
				 control, 
				 order, type) ;
}
void gst_opengl_gluNurbsProperty (OOP nurb, GLenum property, GLfloat value)
{
  gluNurbsProperty (_opengl_vm_proxy->OOPToCObject(nurb), property, value) ;
}
void gst_opengl_gluNurbsSurface (OOP nurb, 
								 GLint sKnotCount, OOP sKnots, /* GLfloat* sKnots, */
								 GLint tKnotCount, OOP tKnots, /* GLfloat* tKnots,  */
								 GLint sStride, GLint tStride, 
								 OOP control, /* GLfloat* control, */
								 GLint sOrder, GLint tOrder, GLenum type)
{
  GLfloat* sKnotsFloat ;
  GLfloat* tKnotsFloat ;
  GLfloat* controlFloat ;
  size_t nObjs = 0 ;
  size_t i ;					/* index */

#ifdef DEBUG
  if( (fp = fopen("/home/olivier/gst.dbg", "a+")) != NULL) {
	fprintf(fp,  "Surface : \n") ;
	fflush(fp) ;
  }
#endif  

  /* Retreive sKnots and put them in an array */
  nObjs = _opengl_vm_proxy->OOPSize(sKnots) ;
  if(nObjs < sKnotCount) 
	return ;					/* Should fire an interrupt */
  sKnotsFloat = alloca (sizeof(GLfloat) * sKnotCount) ;
  for(i = 0 ; i < sKnotCount ; ++i)
	sKnotsFloat[i] = _opengl_vm_proxy->OOPToFloat(_opengl_vm_proxy->OOPAt(sKnots, i)) ;

  /* Retreive tKnots and put them in an array */
  nObjs = _opengl_vm_proxy->OOPSize(tKnots) ;
  if(nObjs < tKnotCount) 
	return ;					/* Should fire an interrupt */
  tKnotsFloat = alloca (sizeof(GLfloat) * tKnotCount) ;
  for(i = 0 ; i < tKnotCount ; ++i)
	tKnotsFloat[i] = _opengl_vm_proxy->OOPToFloat(_opengl_vm_proxy->OOPAt(tKnots, i)) ;
  
  /* Retreive control points and put them in an array */
  nObjs = _opengl_vm_proxy->OOPSize(control) ;
  controlFloat = alloca (sizeof(GLfloat) * nObjs) ;
  for(i = 0 ; i < nObjs ; ++i){
	controlFloat[i] = _opengl_vm_proxy->OOPToFloat(_opengl_vm_proxy->OOPAt(control, i)) ;
  }
  
  
#ifdef DEBUG
  if( fp != NULL) {
	int i ;
	fprintf(fp, "NurbSurface: %x %d (", 
			_opengl_vm_proxy->OOPToCObject(nurb),
			sKnotCount) ;
	for(i=  0 ; i < sKnotCount ; ++i)
	  fprintf(fp, " %f", sKnotsFloat[i]) ;
	fprintf(fp, " ) %d (", tKnotCount) ;
	for(i=  0 ; i < tKnotCount ; ++i)
	  fprintf(fp, " %f", tKnotsFloat[i]) ;
	fprintf(fp, ") %d %d (", sStride, tStride) ;
	for(i=  0 ; i < nObjs ; ++i)	
	  fprintf(fp, " %f", controlFloat[i]) ;
	fprintf(fp, ") %d %d %d\n", sOrder,	tOrder,	type) ;
  }
#endif  

  gluNurbsSurface (_opengl_vm_proxy->OOPToCObject(nurb),
				   sKnotCount,
				   sKnotsFloat,
				   tKnotCount,
				   tKnotsFloat,
				   sStride,
				   tStride,
				   controlFloat,
				   sOrder,
				   tOrder,
				   type) ;

#ifdef DEBUG
  if( fp != NULL) {
	fclose(fp) ;
	fp = NULL ;
  }
#endif  
}

OOP gst_opengl_gluNewNurbsRenderer ()
{
  GLUnurbs* aNurb = gluNewNurbsRenderer() ;
#ifdef DEBUG
  if( (fp = fopen("/home/olivier/gst.dbg", "a+")) != NULL) {
	fprintf(fp,  "Nurb %x : \n", aNurb) ;
	fflush(fp) ;
	fclose(fp) ;
	fp = NULL ;
  }
#endif  
  return _opengl_vm_proxy->cObjectToOOP(aNurb) ;
}

/* Init module */
void gst_init_module_gluNurbs(VMProxy* proxy) {
  int i ;
  _opengl_vm_proxy = proxy ;

#ifdef DEBUG
  if((fp = fopen("/home/olivier/gst.dbg", "a+")) != NULL) {
	fprintf(fp,  "Nurb init\n") ;
	fflush(fp) ;
  }
#endif  
  /* Initialize the Nurbs callback selector and receivers */
  for (i = 0 ; i <NbGluNurbFunctions ; ++i )
	gluNurbReceiver[i] = gluNurbSelector[i] = nil ;

  /* Define C Functions */
  proxy->defineCFunc ("gluNurbsCallback", gst_opengl_gluNurbsCallback) ;
  proxy->defineCFunc ("gluNurbsCallbackData", gst_opengl_gluNurbsCallbackData) ;
  proxy->defineCFunc ("gluNurbsCallbackDataEXT", gst_opengl_gluNurbsCallbackDataEXT) ;
  proxy->defineCFunc ("gluDeleteNurbsRenderer", gst_opengl_gluDeleteNurbsRenderer) ;
  proxy->defineCFunc ("gluGetNurbsProperty", gst_opengl_gluGetNurbsProperty) ;
  proxy->defineCFunc ("gluNurbsCurve", gst_opengl_gluNurbsCurve) ;
  proxy->defineCFunc ("gluNurbsProperty", gst_opengl_gluNurbsProperty) ;
  proxy->defineCFunc ("gluNurbsSurface", gst_opengl_gluNurbsSurface) ;
  proxy->defineCFunc ("gluNewNurbsRenderer", gst_opengl_gluNewNurbsRenderer) ;

#ifdef DEBUG
  if(fp != NULL) {
	fprintf(fp, "Nurbs init Done\n") ;
	fflush(fp) ;
	fclose(fp) ;
	fp = NULL ;
  }
#endif  
}
