#include "config.h"
#include "gstpub.h"

#include <stdio.h>
#include <GL/gl.h>
#include <GL/glut.h>

#define DEBUG 1

#ifdef DEBUG
static FILE* fp = NULL ;
#endif

static VMProxy* _opengl_vm_proxy = NULL ;
#define nil  _opengl_vm_proxy->nilOOP


/* * * * * * * * * */
/* the Tesselator  */
/* callbacks       */
/* * * * * * * * * */



/* Number of known callbacks */
#define NbGluTessFunctions 12
/* Receivers for the callbacks; initialized in module init func */
static OOP gluTessReceiver[NbGluTessFunctions] ;
/* Selectors; initialized in module init func */
static OOP gluTessSelector[NbGluTessFunctions] ;
/* Callback constants definition */
static GLenum gluTessFunctions[NbGluTessFunctions] = {
  GLU_TESS_BEGIN,
  GLU_TESS_VERTEX,
  GLU_TESS_END,
  GLU_TESS_ERROR,
  GLU_TESS_EDGE_FLAG,
  GLU_TESS_COMBINE,
  GLU_TESS_BEGIN_DATA,
  GLU_TESS_VERTEX_DATA,
  GLU_TESS_END_DATA,
  GLU_TESS_ERROR_DATA,
  GLU_TESS_EDGE_FLAG_DATA,
  GLU_TESS_COMBINE_DATA
} ;


/* Retreive the callback index from the Callback contant definition */
int gst_glu_tessCallback_getCallbackIndex(GLenum which)
{
  int i ;
  for (i = 0 ; i < NbGluTessFunctions ; ++i) {
	/* If function is found, return index */
	if (gluTessFunctions[i] == which) {
	  return i ;
	}
  }
  return -1 ;
}

/* Send the callback message to the receiver */
void gst_glu_tessCallback_sendMessage(GLenum which, int nArgs, OOP* args)
{
  /* Get the callback index */
  int index = gst_glu_tessCallback_getCallbackIndex(which) ;
  if(index == -1) return ;
  /* verify the receiver and the selector */
  if(gluTessReceiver[index] != nil && 
	 gluTessSelector[index] != nil) 
	{
	  /* Send the callback message */
	  if(nArgs > 0) 
		_opengl_vm_proxy->nvmsgSend(gluTessReceiver[index], 
									gluTessSelector[index],
									args, nArgs) ;
	  else
		_opengl_vm_proxy->perform(gluTessReceiver[index], 
								  gluTessSelector[index]) ;
	}
}
/* Define the C Callback functions and call the sendMessage function */
void gst_glu_tessCallback_Tess_Begin ( GLenum type )
{
  OOP* args = alloca ( sizeof(OOP) * 1) ;
  args[0] = _opengl_vm_proxy->intToOOP(type) ;
  gst_glu_tessCallback_sendMessage(GLU_TESS_BEGIN, 1, args) ;
}

void gst_glu_tessCallback_Tess_Vertex ( GLfloat* vertex )
{
  OOP* args = alloca ( sizeof(OOP) * 3 ) ;
  args[0] = _opengl_vm_proxy->floatToOOP( vertex[0] ) ;
  args[1] = _opengl_vm_proxy->floatToOOP( vertex[1] ) ;
  args[2] = _opengl_vm_proxy->floatToOOP( vertex[2] ) ;
  gst_glu_tessCallback_sendMessage(GLU_TESS_VERTEX, 3, args) ;
}

void gst_glu_tessCallback_Tess_End ( )
{
  gst_glu_tessCallback_sendMessage(GLU_TESS_END, 0, &nil) ;
}

void gst_glu_tessCallback_Tess_Error (GLenum errorCode)
{
  OOP* args = alloca ( sizeof(OOP) * 1) ;
  args[0] = _opengl_vm_proxy->intToOOP(errorCode) ;
  gst_glu_tessCallback_sendMessage(GLU_TESS_ERROR, 1, args) ;
}

void gst_glu_tessCallback_Tess_EdgeFlag ( int flag ) /* Should be a	GLboolean ! */
{
  OOP* args = alloca ( sizeof(OOP) * 1 ) ;
  args[0] = _opengl_vm_proxy->boolToOOP( flag ) ;
  gst_glu_tessCallback_sendMessage(GLU_TESS_EDGE_FLAG, 1, args) ;
}

void gst_glu_tessCallback_Tess_Combine (GLdouble* coords, void** vertexData, GLfloat* weight, void** outData)
{
  GLdouble* vertex ;
  OOP* args ;
  OOP outDataArray, weightArray, vertexDataArray, coordsArray ;
  int i,j ;
  GLdouble **vd = (GLdouble**) vertexData ;

  args = alloca ( sizeof(OOP) * 4 ) ;

  /*  First arg is a  */
  coordsArray = _opengl_vm_proxy->objectAlloc(_opengl_vm_proxy->arrayClass, 3) ; 
  for(i = 0 ; i < 3 ; ++i)
	_opengl_vm_proxy->OOPAtPut(coordsArray, i,
							   _opengl_vm_proxy->longDoubleToOOP(coords[i])) ;

  /* Then comes 6 vertex datas */
  vertexDataArray = _opengl_vm_proxy->objectAlloc(_opengl_vm_proxy->arrayClass, 24) ;
  for(i = 0 ; i < 6 ; ++i)
	for(j = 0 ; j < 4 ; ++j)
	  _opengl_vm_proxy->OOPAtPut(vertexDataArray, i * 4 + j,
								 _opengl_vm_proxy->longDoubleToOOP(vd[j][i])) ;

  /* Then some weight to apply */
  weightArray = _opengl_vm_proxy->objectAlloc(_opengl_vm_proxy->arrayClass, 4) ;
  for(i = 0 ; i < 4 ; ++i)
	_opengl_vm_proxy->OOPAtPut(weightArray, i * 4 + j,
							   _opengl_vm_proxy->longDoubleToOOP(weight[i])) ;
	
  /* Finaly we should get a returned array of 6 floats */
  outDataArray = _opengl_vm_proxy->objectAlloc(_opengl_vm_proxy->arrayClass, 6) ;

  args[0] = coordsArray ;
  args[1] = vertexDataArray ;
  args[2] = weightArray ;
  args[3] = outDataArray ;

  gst_glu_tessCallback_sendMessage(GLU_TESS_COMBINE, 4, args) ;

  /* Retreive the content of the out data array and assign pointer */
  vertex = alloca(sizeof(GLdouble)*6) ;
  for(i = 0 ; i < 4 ; ++i)
	vertex[i] = _opengl_vm_proxy->OOPToLongDouble(_opengl_vm_proxy->OOPAt(outDataArray, i)) ;
  *outData = vertex ;
}

void gst_glu_tessCallback_Tess_BeginData ( GLenum type, void* data )
{
  OOP* args = alloca ( sizeof(OOP) * 2) ;
  args[0] = _opengl_vm_proxy->intToOOP(type) ;
  args[1] = (OOP) data ;
  gst_glu_tessCallback_sendMessage(GLU_TESS_BEGIN, 2, args) ;
}

void gst_glu_tessCallback_Tess_VertexData ( GLfloat* vertex, void* data )
{
  OOP* args = alloca ( sizeof(OOP) * 4 ) ;
  args[0] = _opengl_vm_proxy->floatToOOP( vertex[0] ) ;
  args[1] = _opengl_vm_proxy->floatToOOP( vertex[1] ) ;
  args[2] = _opengl_vm_proxy->floatToOOP( vertex[2] ) ;
  args[3] = (OOP) data ;
  gst_glu_tessCallback_sendMessage(GLU_TESS_VERTEX, 4, args) ;
}

void gst_glu_tessCallback_Tess_EndData ( void* data )
{
  OOP* args = alloca ( sizeof(OOP) * 1 ) ;
  args[0] = (OOP) data ;
  gst_glu_tessCallback_sendMessage(GLU_TESS_END, 1, args) ;
}

void gst_glu_tessCallback_Tess_ErrorData (GLenum errorCode, void* data)
{
  OOP* args = alloca ( sizeof(OOP) * 2) ;
  args[0] = _opengl_vm_proxy->intToOOP(errorCode) ;
  args[1] = (OOP) data ;
  gst_glu_tessCallback_sendMessage(GLU_TESS_ERROR, 2, args) ;
}

void gst_glu_tessCallback_Tess_EdgeFlagData (int flag, void* data) /* Flag should be a GLboolean ! */
{
  OOP* args = alloca ( sizeof(OOP) * 2 ) ;
  args[0] = _opengl_vm_proxy->boolToOOP( flag ) ;
  args[1] = (OOP) data ;
  gst_glu_tessCallback_sendMessage(GLU_TESS_EDGE_FLAG, 2, args) ;
}

void gst_glu_tessCallback_Tess_CombineData (GLdouble* coords, void** vertexData, GLfloat* weight, void** outData, void* data)
{
  GLdouble* vertex ;
  OOP* args ;
  OOP outDataArray, weightArray, vertexDataArray, coordsArray ;
  int i,j ;
  GLdouble **vd = (GLdouble**) vertexData ;

  args = alloca ( sizeof(OOP) * 5 ) ;

  /*  First arg is a  */
  coordsArray = _opengl_vm_proxy->objectAlloc(_opengl_vm_proxy->arrayClass, 3) ; 
  for(i = 0 ; i < 3 ; ++i)
	_opengl_vm_proxy->OOPAtPut(coordsArray, i,
							   _opengl_vm_proxy->longDoubleToOOP(coords[i])) ;

  /* Then comes 6 vertex datas */
  vertexDataArray = _opengl_vm_proxy->objectAlloc(_opengl_vm_proxy->arrayClass, 24) ;
  for(i = 0 ; i < 6 ; ++i)
	for(j = 0 ; j < 4 ; ++j)
	  _opengl_vm_proxy->OOPAtPut(vertexDataArray, i * 4 + j,
								 _opengl_vm_proxy->longDoubleToOOP(vd[j][i])) ;

  /* Then some weight to apply */
  weightArray = _opengl_vm_proxy->objectAlloc(_opengl_vm_proxy->arrayClass, 4) ;
  for(i = 0 ; i < 4 ; ++i)
	_opengl_vm_proxy->OOPAtPut(weightArray, i * 4 + j,
							   _opengl_vm_proxy->longDoubleToOOP(weight[i])) ;
	
  /* Finaly we should get a returned array of 6 floats */
  outDataArray = _opengl_vm_proxy->objectAlloc(_opengl_vm_proxy->arrayClass, 6) ;

  args[0] = coordsArray ;
  args[1] = vertexDataArray ;
  args[2] = weightArray ;
  args[3] = outDataArray ;
  args[4] = (OOP) data ;

  gst_glu_tessCallback_sendMessage(GLU_TESS_COMBINE, 5, args) ;

  /* Retreive the content of the out data array and assign pointer */
  vertex = alloca(sizeof(GLdouble)*6) ;
  for(i = 0 ; i < 4 ; ++i)
	vertex[i] = _opengl_vm_proxy->OOPToLongDouble(_opengl_vm_proxy->OOPAt(outDataArray, i)) ;
  *outData = vertex ;
}


/* Pointers to the Glu Tess callback functions. It should be exactly */
/* the same order as GlutTessFunctions ! */
static _GLUfuncptr gst_glu_tessCallback[NbGluTessFunctions] = {
  gst_glu_tessCallback_Tess_Begin,
  gst_glu_tessCallback_Tess_Vertex,
  gst_glu_tessCallback_Tess_End,
  gst_glu_tessCallback_Tess_Error,
  gst_glu_tessCallback_Tess_EdgeFlag,
  gst_glu_tessCallback_Tess_Combine,
  gst_glu_tessCallback_Tess_BeginData,
  gst_glu_tessCallback_Tess_VertexData,
  gst_glu_tessCallback_Tess_EndData,
  gst_glu_tessCallback_Tess_ErrorData,
  gst_glu_tessCallback_Tess_EdgeFlagData,
  gst_glu_tessCallback_Tess_CombineData
} ;



void gst_opengl_gluTessCallback (OOP tess, GLenum which, OOP selector, OOP receiver)
{ 
  int index = gst_glu_tessCallback_getCallbackIndex(which) ;
  if(index == -1) 
	return ;

  /* Unregister previous receiver and selector callback */
  if(gluTessReceiver[index] != nil) 
	{
	  _opengl_vm_proxy->unregisterOOP(gluTessReceiver[index]) ;
	  gluTessReceiver[index] = nil ;
	}
  if(gluTessSelector[index] != nil) 
	{
	  _opengl_vm_proxy->unregisterOOP(gluTessSelector[index]) ;
	  gluTessSelector[index] = nil ;
	}
  /* Verify if selector and receiver are not nil. Else clear the callback */
  if(receiver != nil && selector != nil ) 
	{
	  /* Assign new receiver and selector and register them*/
	  gluTessReceiver[index] = receiver ;
	  gluTessSelector[index] = selector ;
	  _opengl_vm_proxy->registerOOP(gluTessReceiver[index]) ;
	  _opengl_vm_proxy->unregisterOOP(gluTessSelector[index]) ;

	  /* Setting up the callback pointer */
	  gluTessCallback (_opengl_vm_proxy->OOPToCObject(tess), which, gst_glu_tessCallback[index]) ;
	}
  else 
	{
	  gluTessReceiver[index] = nil ;
	  gluTessSelector[index] = nil ;

	  /* Cleaning up the callback */
	  gluTessCallback (_opengl_vm_proxy->OOPToCObject(tess), which, NULL) ;
	}
}

/* Init module */
void gst_init_module_gluTess(VMProxy* proxy) {
  int i ;
  _opengl_vm_proxy = proxy ;

#ifdef DEBUG
  fp = fopen("/home/olivier/gst.dbg", "a+") ;
  if(fp != NULL) {
	fprintf(fp,  "Tess init\n") ;
	fflush(fp) ;
  }
#endif  
  /* Initialize the Tess callback selector and receivers */
  for (i = 0 ; i <NbGluTessFunctions ; ++i ) {
	gluTessReceiver[i] = gluTessSelector[i] = nil ;
  }

  /* Define C Functions */
  proxy->defineCFunc ("gluTessCallback", gst_opengl_gluTessCallback) ;
#ifdef DEBUG
  if(fp != NULL) {
	fprintf(fp, "Tess init Done\n") ;
	fflush(fp) ;
	fclose(fp) ;
	fp = NULL ;
  }
#endif  
}
