#include "config.h"
#include "gstpub.h"

#include <stdio.h>
#include <string.h>
#include <GL/glut.h>
#include <stdlib.h>


extern void gst_initModule_glut (VMProxy* proxy) ;
extern void gst_initModule_glu  (VMProxy* proxy) ;
extern void gst_initModule_gl   (VMProxy* proxy) ;


static VMProxy *_opengl_vm_proxy = NULL ;
#define nil  _opengl_vm_proxy->nilOOP

OOP openglClassPointer = NULL ;

/* The number of events is 30. So for now we fix it */
#define MAX_OPENGL_EVENTS 30

OOP openglEventsReceivers[MAX_OPENGL_EVENTS] ;	

#define DEBUG 1

#ifdef DEBUG
static FILE* fp = NULL ;
#endif

static void registerOpenglFuncs (void) ;

void
gst_opengl_connect_signal(OOP receiver, int eventNumber)
{
#ifdef DEBUG
  fp = fopen("/home/olivier/gst.dbg", "a+") ;
  if(fp != NULL) {
	fprintf(fp, "\t\tConnecting %x on event %d\n", receiver, eventNumber) ;
	fflush(fp) ;
  }
#endif  

  /* Get a pointer to OpenGL class to be able to retreive the event number in the callback functions */
  if(! openglClassPointer) {
	registerOpenglFuncs() ;
	openglClassPointer = _opengl_vm_proxy->strMsgSend(receiver, "class", NULL) ;
	if(openglClassPointer != nil ) {
#ifdef DEBUG
  if(fp != NULL) {
	fprintf(fp, "Ok I got class pointer %x\n", openglClassPointer) ;
	fflush(fp) ;
  }
#endif  
	}
  }

  if(eventNumber < 0 || eventNumber >= MAX_OPENGL_EVENTS)
	return ;

  if(openglEventsReceivers[eventNumber]) {
	_opengl_vm_proxy->unregisterOOP(openglEventsReceivers[eventNumber]) ;
#ifdef DEBUG
  if(fp != NULL) {
	fprintf(fp, "Unregister %x\n", openglEventsReceivers[eventNumber]) ;
	fflush(fp) ;
	
  }
#endif  
  }

#ifdef DEBUG
  if(fp != NULL) {
	fprintf(fp, "Registering %x...", receiver) ;
	fflush(fp) ;
  }
#endif  
  _opengl_vm_proxy->registerOOP(receiver) ;
  openglEventsReceivers[eventNumber] = receiver ;
#ifdef DEBUG
  if(fp != NULL) {
	fprintf(fp, "done\n") ;
	fflush(fp) ;
  }
#endif  
}


void
openglSendCallbackMessage(const char* msg, OOP* args, int nArgs) {
  OOP indexOop, callbackList, exec ;
  OOP receiver, selector ;
  int eventNumber ;

#ifdef DEBUG
  if(fp != NULL) {
	fprintf(fp, "\t\tCallback for %s args %x nArgs %d\n", msg, args, nArgs) ;
	fflush(fp) ;
  }
#endif  

  indexOop = _opengl_vm_proxy->strMsgSend(openglClassPointer,
										  msg,
										  NULL) ;
  if(indexOop == nil) return ;

  eventNumber = _opengl_vm_proxy->OOPToInt(indexOop) ;
  
#ifdef DEBUG
  if(fp != NULL) {
	fprintf(fp, "index %d\n", eventNumber) ;
	fflush(fp) ;
  }
#endif  

  if(openglEventsReceivers[eventNumber] == nil)
	return ;

  callbackList = _opengl_vm_proxy->strMsgSend(openglEventsReceivers[eventNumber],
											  "getCallback:",
											  _opengl_vm_proxy->intToOOP(eventNumber),
											  NULL) ;
  if(callbackList == nil)
	return ;

  receiver = _opengl_vm_proxy->strMsgSend(callbackList, 
										  "at:", 
										  _opengl_vm_proxy->intToOOP(1),
										  NULL) ;
  selector = _opengl_vm_proxy->strMsgSend(callbackList, 
										  "at:", 
										  _opengl_vm_proxy->intToOOP(2),
										  NULL) ;

#ifdef DEBUG
  if(fp != NULL) {
	OOP name = _opengl_vm_proxy->strMsgSend(selector, "displayString", NULL) ;
	fprintf(fp, "Receiver %x callback %x(%s)\n", 
			receiver, selector,	_opengl_vm_proxy->OOPToString(name)) ;
	fflush(fp) ;
  }
#endif  

  if(nArgs > 0)
	exec = _opengl_vm_proxy->nvmsgSend(receiver, selector, args, nArgs) ;
  else 
	exec = _opengl_vm_proxy->perform(receiver, selector) ;
	
#ifdef DEBUG
  if(fp != NULL) {
	fprintf(fp, "Done (%x)\n", exec) ;
	fflush(fp) ;
  }
#endif  
}


void 
openglCreateMenu(  int menu  )
{
  OOP* args = alloca( sizeof(OOP) * 1 ) ;
  args[0] = _opengl_vm_proxy->intToOOP(menu) ;

  openglSendCallbackMessage("CreateMenuEvent", args, 1) ;
}

void 
openglTimerFunc( int a )
{
  OOP* args = alloca ( sizeof(OOP) * 1 ) ;
  args[0] = _opengl_vm_proxy->intToOOP(a) ;
  
  openglSendCallbackMessage("TimerFuncEvent", args, 1) ;
}

void 
openglIdleFunc( )
{
  openglSendCallbackMessage("IdleFuncEvent", &nil, 0) ;
}

void 
openglKeyboardFunc(  unsigned char c, int a, int b  )
{
  OOP* args = alloca ( sizeof(OOP) * 3 ) ;
  args[0] = _opengl_vm_proxy->charToOOP(c) ;
  args[1] = _opengl_vm_proxy->intToOOP(a) ;
  args[2] = _opengl_vm_proxy->intToOOP(b) ;
  
  openglSendCallbackMessage("KeyboardFuncEvent", args, 3) ;
}


void 
openglSpecialFunc( int a, int b, int c )
{
  OOP* args = alloca ( sizeof(OOP) * 3 ) ;
  args[0] = _opengl_vm_proxy->intToOOP(a) ;
  args[1] = _opengl_vm_proxy->intToOOP(b) ;
  args[2] = _opengl_vm_proxy->intToOOP(c) ;
  
  openglSendCallbackMessage("SpecialFuncEvent", args, 3) ;
}

void 
openglReshapeFunc( int a, int b )
{
  OOP* args = alloca ( sizeof(OOP) * 2 ) ;
  args[0] = _opengl_vm_proxy->intToOOP(a) ;
  args[1] = _opengl_vm_proxy->intToOOP(b) ;
  
  openglSendCallbackMessage("ReshapeFuncEvent", args, 2) ;
}

void 
openglVisibilityFunc(  int a )
{
  OOP* args = alloca ( sizeof(OOP) * 1 ) ;
  args[0] = _opengl_vm_proxy->intToOOP(a) ;
  openglSendCallbackMessage("VisibilityFuncEvent", args, 1) ;
}

void 
openglDisplayFunc()
{
  OOP* args = NULL ;
  openglSendCallbackMessage("DisplayFuncEvent", args, 0) ;
}

void 
openglMouseFunc(  int a, int b, int c, int d )
{
  OOP* args = alloca ( sizeof(OOP) * 4 ) ;
  args[0] = _opengl_vm_proxy->intToOOP(a) ;
  args[1] = _opengl_vm_proxy->intToOOP(b) ;
  args[2] = _opengl_vm_proxy->intToOOP(c) ;
  args[3] = _opengl_vm_proxy->intToOOP(d) ;
  
  openglSendCallbackMessage("MouseFuncEvent", args, 4) ;
}

void 
openglMotionFunc(  int a, int b )
{
  OOP* args = alloca ( sizeof(OOP) * 2 ) ;
  args[0] = _opengl_vm_proxy->intToOOP(a) ;
  args[1] = _opengl_vm_proxy->intToOOP(b) ;
  
  openglSendCallbackMessage("MotionFuncEvent", args, 2) ;
}

void 
openglPassiveMotionFunc(  int a, int b )
{
  OOP* args = alloca ( sizeof(OOP) * 2 ) ;
  args[0] = _opengl_vm_proxy->intToOOP(a) ;
  args[1] = _opengl_vm_proxy->intToOOP(b) ;
  
  openglSendCallbackMessage("PassiveMotionFuncEvent", args, 2) ;
}

void 
openglEntryFunc(  int a )
{
  OOP* args = alloca ( sizeof(OOP) * 1 ) ;
  args[0] = _opengl_vm_proxy->intToOOP(a) ;

  openglSendCallbackMessage("EntryFuncEvent", args, 1) ;
}

void 
openglKeyboardUpFunc( unsigned char c, int a, int b )
{
  OOP* args = alloca ( sizeof(OOP) * 3 ) ;
  args[0] = _opengl_vm_proxy->charToOOP(c) ;
  args[1] = _opengl_vm_proxy->intToOOP(a) ;
  args[2] = _opengl_vm_proxy->intToOOP(b) ;
  
  openglSendCallbackMessage("KeyboardUpFuncEvent", args, 3) ;
}

void 
openglSpecialUpFunc( int a, int b, int c )
{
  OOP* args = alloca ( sizeof(OOP) * 3 ) ;
  args[0] = _opengl_vm_proxy->intToOOP(a) ;
  args[1] = _opengl_vm_proxy->intToOOP(b) ;
  args[2] = _opengl_vm_proxy->intToOOP(c) ;
  
  openglSendCallbackMessage("SpecialUpFuncEvent", args, 3) ;
}

void 
openglJoystickFunc( unsigned int a, int b, int c, int d )
{
  OOP* args = alloca ( sizeof(OOP) * 4 ) ;
  args[0] = _opengl_vm_proxy->intToOOP(a) ;
  args[1] = _opengl_vm_proxy->intToOOP(b) ;
  args[2] = _opengl_vm_proxy->intToOOP(c) ;
  args[3] = _opengl_vm_proxy->intToOOP(d) ;
  
  openglSendCallbackMessage("JoystickFuncEvent", args, 4) ;
}

void 
openglMenuStateFunc(  int a )
{
  OOP* args = alloca ( sizeof(OOP) * 1 ) ;
  args[0] = _opengl_vm_proxy->intToOOP(a) ;
  
  openglSendCallbackMessage("MenuStateFuncEvent", args, 1) ;
}

void 
openglMenuStatusFunc(  int a, int b, int c )
{
  OOP* args = alloca ( sizeof(OOP) * 3 ) ;
  args[0] = _opengl_vm_proxy->intToOOP(a) ;
  args[1] = _opengl_vm_proxy->intToOOP(b) ;
  args[2] = _opengl_vm_proxy->intToOOP(c) ;
  
  openglSendCallbackMessage("MenuStatusFuncEvent", args, 3) ;
}

void 
openglOverlayDisplayFunc()
{
  
}

void 
openglWindowStatusFunc(  int a )
{
}

void 
openglSpaceballMotionFunc(  int a, int b, int c )
{
  OOP* args = alloca ( sizeof(OOP) * 3 ) ;
  args[0] = _opengl_vm_proxy->intToOOP(a) ;
  args[1] = _opengl_vm_proxy->intToOOP(b) ;
  args[2] = _opengl_vm_proxy->intToOOP(c) ;
  
  openglSendCallbackMessage("MotionFuncEvent", args, 3) ;
}

void 
openglSpaceballRotateFunc(  int a, int b, int c )
{
}

void 
openglSpaceballButtonFunc(  int a, int b )
{
}

void 
openglButtonBoxFunc(  int a, int b )
{
}

void 
openglDialsFunc(  int a, int b )
{
}

void 
openglTabletMotionFunc(  int a, int b )
{
}

void 
openglTabletButtonFunc(  int a, int b, int c, int d )
{
}

void 
openglMouseWheelFunc(  int a, int b, int c, int d)
{
}

void 
openglCloseFunc()
{
}

void 
openglWMCloseFunc()
{
}

void 
openglMenuDestroyFunc()
{
}

void
gst_initModule(proxy)
	 VMProxy *proxy ;
{
  int i ;
  _opengl_vm_proxy = proxy;

  /* Clear the array */
  for(i = 0 ; i <MAX_OPENGL_EVENTS; ++i)
	openglEventsReceivers[i] = nil ;

  proxy->defineCFunc ("openglConnectSignal", gst_opengl_connect_signal) ;


  gst_initModule_glut(_opengl_vm_proxy) ;
  gst_initModule_glu(_opengl_vm_proxy) ;
  gst_initModule_gl(_opengl_vm_proxy) ;
}


void registerOpenglFuncs ()
{
#ifdef DEBUG
  if(fp != NULL) {
	fprintf(fp, "\t\tRegistering callbacks\n") ;
  }
#endif  


  // glutTimerFunc(openglTimerFunc) ;
  //glutIdleFunc(openglIdleFunc) ;
  glutCreateMenu(openglCreateMenu) ;
  glutKeyboardFunc(openglKeyboardFunc) ;
  glutSpecialFunc(openglSpecialFunc) ;
  glutReshapeFunc(openglReshapeFunc) ;
  glutVisibilityFunc(openglVisibilityFunc) ;
  glutDisplayFunc(openglDisplayFunc) ; 
  glutMouseFunc(openglMouseFunc) ;
  glutMotionFunc(openglMotionFunc) ;
  glutPassiveMotionFunc(openglPassiveMotionFunc) ;
  glutEntryFunc(openglEntryFunc) ;
  glutKeyboardUpFunc(openglKeyboardUpFunc) ;
  glutSpecialUpFunc(openglSpecialUpFunc) ;
  // glutJoystickFunc(openglJoystickFunc) ;
  glutMenuStateFunc(openglMenuStateFunc) ;
  glutMenuStatusFunc(openglMenuStatusFunc) ;
  glutOverlayDisplayFunc(openglOverlayDisplayFunc) ;
  glutWindowStatusFunc(openglWindowStatusFunc) ;
  glutSpaceballMotionFunc(openglSpaceballMotionFunc) ;
  glutSpaceballRotateFunc(openglSpaceballRotateFunc) ;
  glutSpaceballButtonFunc(openglSpaceballButtonFunc) ;
  glutButtonBoxFunc(openglButtonBoxFunc) ;
  glutDialsFunc(openglDialsFunc) ;
  glutTabletMotionFunc(openglTabletMotionFunc) ;
  glutTabletButtonFunc(openglTabletButtonFunc) ;

  // glutMouseWheelFunc(openglMouseWheelFunc) ;
  // glutCloseFunc(openglCloseFunc) ;
  // glutWMCloseFunc(openglWMCloseFunc) ;
  // glutMenuDestroyFunc(openglMenuDestroyFunc) ;


}
