PackageLoader fileInPackage: 'OpenGL' !
Namespace current: OPENGL!

Object subclass:#OpenGLTest
	   instanceVariableNames: 'aWindow windowNumber aNurb surface showPoints translate'
	   classVariableNames: ''
	   poolDictionaries: ''
	   category: 'OpenGL' !

!OpenGLTest methodsFor: 'test' !

init
	" Create the window and initialize callbacks "
	showPoints := false.
	translate := Vertex x: 0.0 y: 0.0 z: -5.0.
	aWindow := OpenGL new. 

	aWindow glutInit: 'une surface smalltalkienne'.
	aWindow glutInitDisplayMode: ((OpenGL GlutRgb bitOr: OpenGL GlutDouble) bitOr: OpenGL GlutDepth).
	aWindow glutInitWindowSize: (Point x: 500 y: 500).
	aWindow glutInitWindowPosition: (Point x: 100 y: 100).

	windowNumber := aWindow glutCreateWindow: 'Une surface nurbs' . 
	
	" Init window color and shading model "
	aWindow glClearColor: (Color black).

	aWindow glMaterialfv: OpenGL GlFront
			mode: OpenGL GlDiffuse
			value: (Array with: 0.7
						  with: 0.7
						  with: 0.7
						  with: 1.0).

	aWindow glMaterialfv: OpenGL GlFront
			mode: OpenGL GlSpecular
			value: (Array with: 1.0 
						  with: 1.0 
						  with: 1.0 
						  with: 1.0).
	aWindow glMaterialf: OpenGL GlFront 
			mode: OpenGL GlShininess 
			value: 100.0.

	aWindow glEnable: OpenGL GlLighting. 
	aWindow glEnable: OpenGL GlLight0.
	aWindow glEnable: OpenGL GlDepthTest.
	aWindow glEnable: OpenGL GlAutoNormal.
	aWindow glEnable: OpenGL GlNormalize.
	
	self initSurface.
	
	aNurb := Nurbs newRenderer.
	aNurb property: OpenGL GluSamplingTolerance value: 25.0.
	aNurb property: OpenGL GluDisplayMode value: OpenGL GluFill.
	aNurb addCallback: OpenGL GluError to: self selector: #error:.

	aWindow addCallback: OpenGL KeyboardFuncEvent to: self selector: #keyboard:x:y:.
	aWindow addCallback: OpenGL DisplayFuncEvent to: self selector: #display.
	aWindow addCallback: OpenGL ReshapeFuncEvent to: self selector: #reshape:height:.
	!

initSurface
	| u v |
	"Initializes the control points of the surface to a small hill."
	"The control points range from -3 to +3 in x, y, and z"
	surface := Array new: 48.
	(0 to: 3) do: 
		[:u | (0 to: 3) do: 
				  [:v |
					   surface at: ((u * 12) + (v * 3) + 1) put: (2.0 * (u - 1.5)).
					   surface at: ((u * 12) + (v * 3) + 2) put: (2.0 * (v - 1.5)).
					   ((u = 1 or: [u = 2]) and: [(v = 1) or: [v = 2]])
						   ifTrue: [surface at: ((u * 12) + (v * 3) + 3) put:  3.0 ]
						   ifFalse:[surface at: ((u * 12) + (v * 3) + 3) put: -3.0 ]
						   ]].
!

error: errorCode
	| error |
	error := aWindow gluErrorString: errorCode.
	Transcript show: 'Nurb error : ' , errorCode ; cr.!

renderScene
	aWindow glBegin: OpenGL GlLines.
	aWindow glColor3f: (Color red: 0.8 green: 0.4 blue: 0.4).
	aWindow glVertex3f: -10.0 y: 0.0 z: 0.0.
	aWindow glVertex3f:  10.0 y: 0.0 z: 0.0.
	aWindow glColor3f: (Color red: 0.4 green: 0.8 blue: 0.4).
	aWindow glVertex3f: 0.0 y: -10.0 z: 0.0.
	aWindow glVertex3f: 0.0 y:  10.0 z: 0.0.
	aWindow glColor3f: (Color red: 0.4 green: 0.4 blue: 0.8).
	aWindow glVertex3f: 0.0 y: -10.0 z: 0.0.
	aWindow glVertex3f: 0.0 y:  10.0 z: 0.0.
	aWindow glEnd.
!

display
	| knots i j |
	knots := #(0.0 0.0 0.0 0.0 1.0 1.0 1.0 1.0).
	aWindow glClear: (OpenGL GlColorBufferBit bitOr: OpenGL GlDepthBufferBit).
	aWindow glPushMatrix.
	aWindow glRotatef: 330.0 direction: (Vertex x: 1.0 y: 0.0 z: 0.0).
	aWindow glScalef: (Vertex x: 0.5 y: 0.5 z: 0.5).
	aWindow gluBeginSurface: aNurb.
	aNurb surface: 8 sKnots: knots tKnotCounts: 8 tKnots: knots
		  sStride: (4 * 3) tStride: 3 control: surface
		  sOrder: 4 tOrder: 4
		  type: OpenGL GlMap2Vertex3 .
	aWindow gluEndSurface: aNurb.

	showPoints ifTrue: [
		aWindow glPointSize: 5.0.
		aWindow glDisable: OpenGL GlLighting.
		aWindow glColor3f: (Color red: 1.0 green: 1.0 blue: 0.0).
		aWindow glBegin: OpenGL GlPoints.
		(0 to: 3) do: [:i |
			(0 to: 3) do: [:j |
				aWindow
					glVertex3f: (surface at: (i*12)+(j*3)+1)
					y: (surface at: (i*12)+(j*3)+2)
					z: (surface at: (i*12)+(j*3)+3).
				].
			].

		aWindow glEnd.
		aWindow glEnable: OpenGL GlLighting.
		].
	aWindow glPopMatrix.
	aWindow glutSwapBuffers.
	!

reshape: w height: h
	aWindow glViewport: (Point x: 0 y: 0) extend: (Point x: w y: h).
	aWindow glMatrixMode: OpenGL GlProjection.
	aWindow glLoadIdentity.
 	aWindow gluPerspective: 45.0 aspect: (1.0*w/h) near: 3.0 far: 8.0.
	aWindow glMatrixMode: OpenGL GlModelview. 
	aWindow glLoadIdentity.
	aWindow glTranslatef: translate.
	!

keyboard: aKey x: aX y: aY
	aKey = $c 
		ifTrue:[
			showPoints := showPoints not.
			aWindow glutPostRedisplay.
			Transcript show: 'Points : ' , showPoints printString.
			].
	aKey = $x ifTrue: [translate x: (translate x + 0.5)].
	aKey = $X ifTrue: [translate x: (translate x - 0.5)].
	aKey = $y ifTrue: [translate y: (translate y + 0.5)].
	aKey = $Y ifTrue: [translate y: (translate y - 0.5)].
	aKey = $z ifTrue: [translate z: (translate z + 0.5)].
	aKey = $Z ifTrue: [translate z: (translate z - 0.5)].
	aWindow glLoadIdentity.
	aWindow glTranslatef: translate.
	aWindow glutPostRedisplay.
	!
!


!OpenGLTest methodsFor: 'access' !
window
	^aWindow !
window: a
	aWindow := a! 
!

OpenGLTest new init.
OpenGL mainIteration. !
